package events;

import communication.Comm;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import view.GraphicUI;
import java.util.HashSet;
import java.util.Set;
import webcam.Snapshot;

/**
 * Reseni stisku klavesy na klavesnici. Resi logovani v vola logovani v GUI
 * a odesilani v komunikacni tride pri stisku prislusne klavesy.
 *
 * @author Michael
 */
public class KeyboardListener implements KeyListener {
    
    private static final int POCET_PACKETU = 8; //pocet packetu pro otoceni o 90°

    private Comm comm;
    private GraphicUI gui;
    private final Set<Integer> pressed = new HashSet<>();

    /**
     * Vlasni konstruktor pro zavedeni nutnych objektu.
     * @param g Graficke prostredi.
     * @param c Komunikacni trida.
     */
    public KeyboardListener(GraphicUI g, Comm c) {
        gui = g;
        comm = c;
    }

    /**
     * Zmacklou klavesu vlozi do Setu a zavola metodu rozhodujici o nasledne
     * akci.
     * @param e
     */
    @Override
    public void keyPressed(KeyEvent e) {
        if (pressed.size() < 3) {
            pressed.add(e.getKeyCode());
        }

        reactToKey();
    }

    /**
     * Nepotrebna, ale nutna metoda z KeyListeneru.
     * @param e
     */
    @Override
    public void keyTyped(KeyEvent e) {}

    /**
     * Akce pri pusteni jakekoli klavesy.
     * Nastavuje vizualizacni sipky s GUI na puvodni podobu.
     * @param e
     */
    @Override
    public void keyReleased(KeyEvent e) {
        try {
            pressed.remove(e.getKeyCode());
            
            gui.setArrowClicked(0, false);
            gui.setArrowClicked(1, false);
            gui.setArrowClicked(2, false);
            gui.setArrowClicked(3, false);
            if(pressed.size() == 0) {
                comm.sendPacket(0);
            } else {
                reactToKey();
            }
        } catch (Exception ex) {
            gui.logEvents("Send packet err");
        }
    }

    /**
     * Projde Set a vrati prikaz k odeslani podle zmacklych klaves.
     * @return Prikaz k odeslani.
     */
    private int getKeyCode() {
        int key = 0;

        if(pressed.size() > 1) {
            // dve stisknute klavesy
            if(pressed.contains(KeyEvent.VK_UP) && pressed.contains(KeyEvent.VK_RIGHT))
                key = 2;
            else if(pressed.contains(KeyEvent.VK_DOWN) && pressed.contains(KeyEvent.VK_RIGHT))
                key = 4;
            else if(pressed.contains(KeyEvent.VK_DOWN) && pressed.contains(KeyEvent.VK_LEFT))
                key = 6;
            else if(pressed.contains(KeyEvent.VK_UP) && pressed.contains(KeyEvent.VK_LEFT))
                key = 8;
        } else {
            // jedna stisknuta klavesa
            if(pressed.contains(KeyEvent.VK_SPACE))
                key = 10;
            else if (pressed.contains(KeyEvent.VK_ENTER))
                key = 11;
            else if(pressed.contains(KeyEvent.VK_UP))
                key = 1;
            else if(pressed.contains(KeyEvent.VK_RIGHT))
                key = 3;
            else if(pressed.contains(KeyEvent.VK_DOWN))
                key = 5;
            else if(pressed.contains(KeyEvent.VK_LEFT))
                key = 7;
        }

        return key;
    }

    /**
     * Definovany reakce na zmackle jednotlive klavesy.
     * up = 1;
     * up+right = 2;
     * right = 3;
     * down+right = 4;
     * down = 5;
     * down+left = 6;
     * left = 7;
     * up+left = 8;
     * space = 10 (stop);
     * enter = 11 (center along road);
     */
    private void reactToKey() {
        try {
            switch (getKeyCode()) {
                case 1:
                    gui.setArrowClicked(0, true);
                    comm.sendPacket(1);
                    gui.logEvents("up");
                    break;
                case 2:
                    gui.setArrowClicked(0, true);
                    gui.setArrowClicked(1, true);
                    comm.sendPacket(2);
                    gui.logEvents("up + right");
                    break;
                case 3:
                    gui.setArrowClicked(1, true);
                    comm.sendPacket(3);
                    gui.logEvents("right");
                    break;
                case 4:
                    gui.setArrowClicked(1, true);
                    gui.setArrowClicked(2, true);
                    comm.sendPacket(4);
                    gui.logEvents("down + right");
                    break;
                case 5:
                    gui.setArrowClicked(2, true);
                    comm.sendPacket(5);
                    gui.logEvents("down");
                    break;
                case 6:
                    gui.setArrowClicked(2, true);
                    gui.setArrowClicked(3, true);
                    comm.sendPacket(6);
                    gui.logEvents("down + left");
                    break;
                case 7:
                    gui.setArrowClicked(3, true);
                    comm.sendPacket(7);
                    gui.logEvents("left");
                    break;
                case 8:
                    gui.setArrowClicked(0, true);
                    gui.setArrowClicked(3, true);
                    comm.sendPacket(8);
                    gui.logEvents("up + left");
                    break;
                case 10:
                    comm.sendPacket(10);
                    gui.logEvents("space");
                    break;
                case 11:
                    centerAtRoad();
                    gui.logEvents("enter");
                    break;
            }
        } catch(Exception io) {
            gui.logEvents("Send packet err");
        }
    }
    
    /**
     * Zjisti si divergence (odklon robota od idelani cesty) a srovna jej.
     * @throws IOException 
     */
    private void centerAtRoad() throws IOException {
        float div = Snapshot.getDivergence();
        int kolikrat = Math.round(((float)POCET_PACKETU) * Math.abs(div));
        
        if(div == 0);
        else if(div < 0) {
            for(int i = 0; i < kolikrat; i++) {
                comm.sendPacket(7);
            }
        } else {
            for(int i = 0; i < kolikrat; i++) {
                comm.sendPacket(3);
            }
        }
    }

}