package imageProcessing;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import javax.imageio.ImageIO;

/**
 * Hlavni trida pro nalezeni tvaru v obrazku a analyzu cesty, na ktere je robot.
 * 
 * EdgeProcessor ep = new EdgeProcessor();
 * 
 * try {
 *     ep.findEdges();
 * } catch (IOException ex) {
 *     System.out.println("Image load error");
 * }
 *
 * ep.analyzePath();
 * System.out.println(ep.isLeftClear() + " < L|R > " + ep.isRightClear());
 * System.out.println((ep.returnDivergence()*100) + " %");
 *  
 * ArrayList alS = ep.findShapes();
 * 
 * @author Michael
 */
public class ImageProcessor {
    
    private static final float LOW_TRESH = 3.0f; //spodni treshold pro hledani hran
    private static final float HIGH_TRESH = 7.0f; //horni treshold pro hledani hran
    
    private static CannyEdgeDetector detector;
    private int horizonatalLimit; //limit pro zjisteni, zda je robot miri mimo cestu
    private int boundEdgeTop; //horni okraj vyrezu (omezeni zjistovane oblasti)
    private int boundEdgeBottom; //dolni okraj vyrezu (omezeni zjistovane oblasti)
    
    private String pathToOriImage; //cesta k souboru obrazku pro zkoumani
    private int height;
    private int width;
    private BufferedImage edgeImage;
    
    private int[] distanceFromMid;
    
    /**
     * Konstruktor.
     * Nastavi objekt, kdy pocatecni hodnoty jsou:
     * cesta(img/snapshot.jpg), horizontalLimit(200px), boundEdgeTop(460px),
     * boundEdgeBottom(465px).
     */
    public ImageProcessor() {
        pathToOriImage = "img/snapshot.jpg";
        edgeImage = null;
        height = 0;
        width = 0;
        horizonatalLimit = 200;
        boundEdgeTop = 460;
        boundEdgeBottom = 465;
        distanceFromMid = new int[2];
        
        detector = new CannyEdgeDetector();
        detector.setLowThreshold(LOW_TRESH);
        detector.setHighThreshold(HIGH_TRESH);
    }
    
    /**
     * Konstruktor.
     * Dela vse, co zakladni, ale nastavuje jinou cestu k souboru.
     * @param path Cesta k souboru zkoumaneho obrazku (napr. "img/snapshot.jpg")
     */
    public ImageProcessor(String path) {
        super();
        pathToOriImage = path;
    }
    
// GETs/SETs
    
    /**
     * Vrati obrazek nalezenych hran.
     * @return Obrazek hran, nebo null pokud nebyl jeste vytvoren.
     */
    public BufferedImage getEdgesImage() {
        return edgeImage;
    }
    
    /**
     * Vrati nastaveny horizontalni limit pro nabourani do hrany cesty.
     * @return Limit v pixelech od stredu snimku
     */
    public int getHorizonatalLimit() {
        return horizonatalLimit;
    }
    
    /**
     * Vrati nastavenou horni a dolni hranu vyrezu z obrazku.
     * @return Pole 2 hodnot {horni hrana, dolni hrana}. Cisla v pixelech od
     * vrchniho leveho okraje.
     */
    public int[] getVerticalBounds() {
        int[] a = {boundEdgeTop,boundEdgeBottom};
        return a;
    }
    
    /**
     * Vrati cestu k zdrojovemu souboru, ktery se zkouma.
     * @return Cesta
     */
    public String getPathToImage() {
        return pathToOriImage;
    }
    
    /**
     * Nastavi obrazek s hranama.
     * @param image 
     */
    public void setEdgesImage(BufferedImage image) {
        edgeImage = image;
    }
    
    /**
     * Nastavi cestu k souboru obrazku urcenemu k analyze.
     * @param path 
     */
    public void setPathToImage(String path) {
        pathToOriImage = path;
    }
    
    /**
     * Nastavi horizontalni limit. Ktery je urcen k zjistovani, zda je robot na
     * ceste nebo uz mimo ni. Je to vzdalenost od stredu zkoumaneho obrazku.
     * @param limit Maximalne polovina sirky zkoumaneho obrazku.
     */
    public void setHorizontalLimit(int limit) {
        horizonatalLimit = limit;
    }
    
    /**
     * Nastavi horni a dolni hranu orezu zkoumane oblasti. Podle techto limitu
     * se oreze obrazek, ktery se pak teprve analyzuje. Vzdalenost v pixelech od
     * horniho leveho okraje.
     * @param top Horni limit
     * @param bottom Dolni limit
     */
    public void setVerticalLimits(int top, int bottom) {
        boundEdgeTop = top;
        boundEdgeBottom = bottom;
    }
    
// public
    
    /**
     * Analyzuje cestu pro nasledne zjistovani, zda ma robot cestu "otevrenou",
     * nebo je spatne natocen a o kolik. Nutno provest po "findEdges" a pred
     * vyuzitim "isLeftClear","isRightClear","returnDivergence".
     */
    public void analyzePath() {
        getDistanceFromMid(cropPixels(getPixels()));
    }
    
    /**
     * Zjisti, zda je cesta vlevo od robota volna, nebo naboura.
     * Musi predchazet "findEdges" a "analyzePath".
     * @return True pokud je volna
     */
    public boolean isLeftClear() {
        return ((distanceFromMid[0] != (-1)) && (distanceFromMid[0] < horizonatalLimit)) ? false : true;
    }
    
    /**
     * Zjisti, zda je cesta vpravo od robota volna, nebo naboura.
     * Musi predchazet "findEdges" a "analyzePath".
     * @return True pokud je volna
     */
    public boolean isRightClear() {
        return ((distanceFromMid[1] != (-1)) && (distanceFromMid[1] < horizonatalLimit)) ? false : true;
    }
    
    /**
     * Vrati o kolik je robot vychylen z primeho kurzu mezi hranami cesty.
     * (-1 = moc vlevo | 0 = ideal | 1 = moc vpravo). Musi predchazet
     * "findEdges" a "analyzePath".
     * @return Cislo mezi -1 az 1, 0 je idealni stred.
     */
    public float returnDivergence() {
        float l = (float) distanceFromMid[0];
        float r = (float) distanceFromMid[1];
        return ((r-l)/(r+l));
    }
    
    /**
     * Analyzuje zkoumany obrazek a pripravi se na vyzadani vysledku.
     * @return Vysledny obrazek hran (neni nutno vyuzit)
     * @throws IOException Chyba u nacteni souboru ze zadane cesty
     */
    public BufferedImage findEdges() throws IOException {
        BufferedImage image = ImageIO.read(new File(pathToOriImage));
        height = image.getHeight();
        width = image.getWidth();
        
        detector.setSourceImage(image);
        detector.process();
        edgeImage = detector.getEdgesImage();
        
        return edgeImage;
    }
    
    /**
     * Analyzuje zkoumany obrazek a pripravi se na vyzadani vysledku.
     * Nenacita soubor z disku, ale analyzuje primo vlozeny.
     * @param originalImage Obrazek pro zkoumani
     * @return Vysledny obrazek hran (neni nutno vyuzit)
     */
    public BufferedImage findEdges(BufferedImage originalImage) {
        height = originalImage.getHeight();
        width = originalImage.getWidth();
        
        detector.setSourceImage(originalImage);
        detector.process();
        edgeImage = detector.getEdgesImage();
        
        return edgeImage;
    }
    
    /**
     * Nalezne tvary ve vyrezu (pro zrychleni) zanalyzovaneho obrazku.
     * Musi predchazet metoda "findEdges".
     * @return Seznam tvaru z obrazku
     */
    public ArrayList<Shape> findShapes() {
        ShapeFinder sp = new ShapeFinder(boundEdgeBottom-boundEdgeTop, width);
        
        return sp.findShapes(cropPixels(getPixels()));
    }
    
    /**
     * Zapise na disk soubor z obrazku nalezenych hran.
     * Cesta k souboru bude naprosto totozna s originalem a nazev souboru bude:
     * "*_e.gif". Musi predchazet "findEdges".
     * @throws IOException Chyba pri zapisu souboru
     */
    public void writeEdgeFile() throws IOException {
        ImageIO.write(edgeImage, "gif", new File(pathToOriImage + "_e.gif"));
    }
    
// private
    
    /**
     * Vrati pole pixelu z obrazku hran.
     * @return Pole pixelu
     */
    private int[] getPixels() {
        return (int[]) edgeImage.getData().getDataElements(0, 0, width, height, null);
    }
    
    /**
     * Orizne pole pixelu podle "boundEdgeTop", "boundEdgeBottom" a prevede
     * barvu pixelu do rozsahu 0 az 255.
     * @param pixels Pole originalnich pixelu obrazku hran
     * @return Pole pixelu vyrezu
     */
    private short[] cropPixels(int[] pixels) {
        int i = boundEdgeTop * width;
        int end = boundEdgeBottom * width;
        short[] pix = new short[end - i];
        
        for(int j = 0; i < end; i++,j++){
            pix[j] = (short)(pixels[i] & 0xff);
        }
        
        return pix;
    }
    
    /**
     * Propocita pixely hran cesty vlevo a vpravo od stredu vyriznuteho obrazku
     * a pripravi je k dalsi analyze.
     * @param pix Pole pixelu oriznuteho obrazku
     */
    private void getDistanceFromMid(short[] pix) {
        int sumaL = 0;
        int sumaR = 0;
        int pocetL = 0;
        int pocetR = 0;
        int stred = Math.round((float)width/(float)2);
        
        for(int i = 0; i < pix.length; i++) {
            int x = i % width;
            if(pix[i] > 0) {
                if(x < stred) {
                    sumaL += stred - x;
                    pocetL++;
                } else {
                    sumaR += x - stred;
                    pocetR++;
                }
            }
        }
        
        distanceFromMid[0] = (pocetL > 0) ? (sumaL/pocetL) : (stred);
        distanceFromMid[1] = (pocetR > 0) ? (sumaR/pocetR) : (stred);
    }
    
}