package imageProcessing;

import java.util.ArrayList;

/**
 * Provadi veskere operace spojene s nalezenim tvaru v obrazku.
 * Obrazek musi byt ve forme pole cernobilych pixelu, kdy bile jsou hrany tvaru
 * a cerna je pro zbytek.
 *
 * @author Michael
 */
public class ShapeFinder {
    
    private static final int OKOLI_BODU = 1; //pocet policek okolo bodu pro vypocet souradnice
    private static final float TOLERANCE_PRO_TVAR = 2.0f; //pod tuto delku se mazou cary
    
    private ArrayList<Shape> tvary;
    private ArrayList<Line> cary;
    private int height;
    private int width;
    
    private short[] pixely;
    
    
    /**
     * Konstruktor. Nastavi pocatecni hodnoty.
     * @param height Vyska testovaneho obrazku
     * @param width Sirka testovaneho obrazku
     */
    public ShapeFinder(int height, int width) {
        tvary = new ArrayList<>();
        cary = new ArrayList<>();
        this.height = height;
        this.width = width;
    }
    
// GETs/SETs
    
    /**
     * Vrati nastavenou vysku.
     * @return Vyska
     */
    public int getHeight() {
        return height;
    }
    
    /**
     * Vrati nastavenou sirku.
     * @return Sirka
     */
    public int getWidht() {
        return width;
    }
    
    /**
     * Nastavi vysku a sirku pozdeji zkoumaneho obrazku.
     * @param h Vyska
     * @param w Sirka
     */
    public void setHeightWidht(int h, int w) {
        height = h;
        width = w;
    }
    
// public
    
    /**
     * Provede veskere operace spojene s nalezenim tvaru v obrazku.
     * @param polePixelu Pole pixelu zkoumaneho obrazku
     * @return Seznam tvaru ze zkoumaneho obrazku
     */
    public ArrayList<Shape> findShapes(short[] polePixelu) {
        cary.clear();
        tvary.clear();
        pixely = polePixelu;
        
        //sestavi cary
        for(int i = 0; i < pixely.length; i++) {
            if(pixely[i] > 0) {
                addPointToLine(i);
            }
        }
        
        //sestavi tvary
        constructShapes();
        
        return tvary;
    }
    
// private
    
    /**
     * Sestavi tvary z nalezenych usecek v obraze, pokud nelze usecka nikomu pridat,
     * pak pro ni vytvori novy tvar. Po pridani do tvaru je cara smazana ze seznamu car.
     */
    private void constructShapes() {
        //promazani prilis kratkych car
        for(int i = 0; i < cary.size();) {
            Line l = cary.get(i);
            if(l.getFirstPoint().distanceFrom(l.getLastPoint()) < TOLERANCE_PRO_TVAR) {
                cary.remove(i);
                i = 0;
            } else {
                l.countABC();
                i++;
            }
        }
        //ini prvni
        if(!cary.isEmpty()) {
            tvary.add(new Shape(cary.remove(0)));
        }
        //rozdeli cary tvarum
        int nShapes = 0;
        int indexCary = 0;
        while(!cary.isEmpty()) {
            if(indexCary < cary.size()) {
                if(tvary.get(nShapes).addLine(cary.get(indexCary))) {
                    cary.remove(indexCary);
                    indexCary = 0;
                } else {
                    indexCary++;
                }
            } else {
                indexCary = 0;
                tvary.add(new Shape(cary.remove(indexCary)));
                nShapes++;
            }
        }
        //promaze nevyhovujici tvary
        for(int j = 0; j < tvary.size(); j++){
            Shape s = tvary.get(j);
            int i = 0;
            while(i < s.getLines().size()) {
                Line l = s.getLines().get(i);
                if(l.getFirstPoint().distanceFrom(l.getLastPoint()) < TOLERANCE_PRO_TVAR) {
                    s.getLines().remove(i);
                    i = 0;
                } else {
                    i++;
                }
            }
            if(i == 0) {
                tvary.remove(j);
            }
        }
    }
    
    /**
     * Hleda ktere care by bod pridal, kdyz se nepovede, pak vytvori novou pro
     * dany bod.
     * @param pixelPos Index pixelu z pole pixelu
     */
    private void addPointToLine(int pixelPos) {
        Point p = computePoint(pixelPos);
        
        if(cary.isEmpty()) {
            cary.add(new Line(p));
        } else {
            boolean isAdded = false;
            for(Line l : cary) {
                if(l.addPoint(p)) {
                    isAdded = true;
                    break;
                }
            }
            if(!isAdded) {
                cary.add(new Line(p));
            }
        }
    }
    
    /**
     * Spocita bod ze zadaneho pixelu a jeho blizkeho okoli omezeneho parametrem
     * "OKOLI_BODU".
     * @param pixelPos Index prostredniho pixelu z pole pixelu
     * @return Vypocitany bod
     */
    private Point computePoint(int pixelPos) {
        ArrayList<Point> barevnePixely = new ArrayList<>();
        
        for(int rad = 0-OKOLI_BODU; rad < OKOLI_BODU+1; rad++) {
            for(int slo = 0-OKOLI_BODU; slo < OKOLI_BODU+1; slo++) {
                if(isSomethingOnPoint(pixelPos + (rad * width) + slo)) {
                    barevnePixely.add(new Point( (pixelPos%width) + slo , (pixelPos/width) + rad ));
                }
            }
        }
        
        int sumX = 0;
        int sumY = 0;
        
        for (Point bod : barevnePixely) {
            sumX += bod.getX();
            sumY += bod.getY();
        }
        
        return new Point((float)sumX/(float)barevnePixely.size(), (float)sumY/(float)barevnePixely.size());
    }
    
    /**
     * Testuje zda neni pozice pixelu mimo pole a jestli pixel obsahuje bod hrany.
     * @param pixelPos Index pixelu z pole pixelu
     * @return True pokud je bod v poli a obsahuje bod hrany
     */
    private boolean isSomethingOnPoint(int pixelPos) {
        if((pixelPos < 0) || (pixelPos > (pixely.length - 1))) {
            return false;
        } else {
            if(pixely[pixelPos] > 0) {
                pixely[pixelPos] = 0;
                return true;
            } else {
                return false;
            }
        }
    }
    
}
